#include "gnome_func_hint.h"
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <gdk/gdkx.h>

#include "sharedm.h" /* shared memory functions */

static GdkAtom _XA_WIN_SIGNAL_FUNC_HINT;      /* the request atom */
static GdkAtom _XA_WIN_SIGNAL_FUNC_HINT_BACK; 

static GList *list;
static GList *llist;

static gboolean hint_need_init = TRUE;

typedef enum _hint
{
  LISTENER,
  CLIENT
}Hint;

typedef struct _GnomeSignalFuncCallback
{
  Hint who;            
  GtkSignalFunc func;  
  GtkWidget *widget;  
  Window xwindow;    
  gpointer data;
  gchar *name;
  gint32 key;
}GnomeSignalFuncCallback;


static void gnome_signal_hint_event_handler( GtkWidget *widget, GdkEvent *event, gpointer data ); 
static void gnome_func_hint_callback( GtkWidget *widget, gpointer data );

void gnome_signal_func_hint_init()
{
   g_return_if_fail( GDK_DISPLAY() );
  _XA_WIN_SIGNAL_FUNC_HINT = XInternAtom( GDK_DISPLAY(), XA_WIN_SIGNAL_FUNC_HINT, False );
  _XA_WIN_SIGNAL_FUNC_HINT_BACK = XInternAtom( GDK_DISPLAY(), XA_WIN_SIGNAL_FUNC_HINT_BACK, False );
  list = NULL;
  llist = NULL;
  hint_need_init = FALSE;
}

/* client message to the listeners */
gboolean gnome_signal_func_hint_set( GtkWidget *widget, gchar* name, GtkSignalFunc func, gpointer data )
{
  GdkEvent ev;
  GnomeSignalFuncCallback *call;
  GdkWindowPrivate *priv; 
  gint size;

  if( GTK_IS_OBJECT( widget ) == FALSE )
    return FALSE;

  if( hint_need_init )
    {
      gnome_signal_func_hint_init();
      gtk_signal_connect( GTK_OBJECT( widget ), "event",
			  GTK_SIGNAL_FUNC( gnome_signal_hint_event_handler ), (gpointer) CLIENT );
    }

  priv = (GdkWindowPrivate*)(GTK_WIDGET(widget)->window);

  call = (GnomeSignalFuncCallback*) g_malloc( sizeof(GnomeSignalFuncCallback));

  call->who = CLIENT;
  call->widget = widget;
  call->xwindow = priv->xwindow;
  call->func = func;
  call->data = data;
  call->name = g_strdup( name );
  size = strlen( call->name ) + 1; /* don't forget the '\0' character! */
/*  g_print( "Size: %d \tName: %s\n", size, call->name );*/
  call->key = gnome_shared_mem_write( call->name, size);

  list = g_list_append( list, call );

  ev.type = GDK_CLIENT_EVENT;
  ev.client.type = GDK_CLIENT_EVENT;
  ev.client.window = widget->window;
  ev.client.send_event = TRUE;
  ev.client.message_type = _XA_WIN_SIGNAL_FUNC_HINT;
  ev.client.data_format = 32;
  ev.client.data.l[0] = (glong) priv->xwindow;  /* integer identifying the xwindow */
  ev.client.data.l[1] = (glong) call->key;      /* shared memory key */
  ev.client.data.l[2] = (glong) size;
    

 /*  register funtion at listeners */
  gdk_event_send_clientmessage_toall( &ev );   /* send messafe to all toplevel windows */

  return TRUE;
}

/* event handler for listening to the hints */
static void gnome_signal_hint_event_handler( GtkWidget *widget, GdkEvent *event, gpointer data )
{
  GtkSignalFunc func;
  GnomeSignalFuncListener *listener;
  GnomeSignalFuncCallback *call = (GnomeSignalFuncCallback*) data;
  Hint hint;
  GList *node;

  hint = (Hint) data;

 if( hint_need_init )
    gnome_signal_func_hint_init();

  switch( event->type )
    {
    case GDK_ENTER_NOTIFY:
    case GDK_LEAVE_NOTIFY:
      break;
    case GDK_CLIENT_EVENT:
      if( hint == LISTENER )
	{
	  if( event->client.message_type == _XA_WIN_SIGNAL_FUNC_HINT )
	    { 
	      for( node=list; node; node = node->next ) /* searching for the right callback */
		{
		  call = (GnomeSignalFuncCallback*) node->data;

		  listener = (GnomeSignalFuncListener*) g_malloc( sizeof(GnomeSignalFuncListener) );
		  listener->strtitel = (gchar*) gnome_shared_mem_read( event->client.data.l[1], event->client.data.l[2] );
		  listener->func = (GtkSignalFunc) gnome_func_hint_callback;
		  listener->xwindow = event->client.data.l[0];
		  listener->key = event->client.data.l[1];
		  listener->data = call->data;
		  call->key = event->client.data.l[1];
		  func = call->func;
		  func( call->widget, listener );  /* call the register function of the listener */

		  llist = g_list_append( llist, listener ); /* against memory wholes */
		}
	    }
	}
      else if( hint == CLIENT )
	{
	  if( event->client.message_type == _XA_WIN_SIGNAL_FUNC_HINT_BACK )
	    {
	      for( node=list; node; node=node->next ) /* searching for the right client callback */
		{
		  call = (GnomeSignalFuncCallback*) node->data;
		  if( (call->key == event->client.data.l[1]) && (call->who == CLIENT ))
		    {
		      func = call->func;
		      func( call->widget, call->data );
		      break;
		    }
		}
	    } 
	}
      break;
    default:
      break;
    }
}

/* register the listener function, which is listening to the client messages */
gboolean gnome_signal_func_hint_get( GtkSignalFunc func, GtkWidget *widget, gpointer data )
{ 
  GnomeSignalFuncCallback *call;

  if( GTK_IS_OBJECT( widget ) == FALSE )
    return FALSE;
  

 if( hint_need_init )
   {
     gnome_signal_func_hint_init();
     gtk_signal_connect( GTK_OBJECT( widget ), "event",
			 GTK_SIGNAL_FUNC( gnome_signal_hint_event_handler ), (gpointer) LISTENER );
   }

 call =(GnomeSignalFuncCallback*) g_malloc( sizeof(GnomeSignalFuncCallback) );
 
 call->who = LISTENER;
 call->widget = widget;
 call->func = func;
 call->data = data;
 
 list = g_list_append( list, call ); 
}

/* the function which the listener has to call.
 * the data pointer have to be a pointer on the 
 * GnomeSignalFuncListener structure which was sent to 
 * listen function of the listener !!!!
 */
static void gnome_func_hint_callback( GtkWidget *widget, gpointer data )
{
  GnomeSignalFuncListener *listener;
  GnomeSignalFuncCallback *call;
  GdkEvent ev;
  GList *node;

  if( hint_need_init )
    gnome_signal_func_hint_init();
    
  listener = (GnomeSignalFuncListener*) data;

  ev.type = GDK_CLIENT_EVENT;
  ev.client.type = GDK_CLIENT_EVENT;
  ev.client.send_event = TRUE;
  ev.client.message_type = _XA_WIN_SIGNAL_FUNC_HINT_BACK;
  ev.client.data_format = 32;
  ev.client.data.l[0] = listener->xwindow;
  ev.client.data.l[1] = listener->key;
  
  /* send it back */
  gdk_event_send_client_message( &ev, listener->xwindow );  
}

/* free all */
void gnome_signal_func_hint_quit()
{
  GList *node;
  GnomeSignalFuncCallback *call;
  GnomeSignalFuncListener *listener;

  for( node = list; node; node = node->next )
    {
      call = (GnomeSignalFuncCallback*) node->data;

      if( call->who = CLIENT )
	{
	  g_free( call->name );
/*	  g_print( "Key: %ld freed\n", (long) call->key ); */
	  gnome_shared_mem_free( call->key );
	  g_free( call );
	}
      else if( call->who == LISTENER )
	{
	  g_free( call );
	}
    }
  g_list_free( list );

  for( node = llist; node; node = node->next )
    {
      listener = (GnomeSignalFuncListener*) node->data;
      g_free( listener->strtitel );
      g_free( listener );
    }
  g_list_free( llist );

  llist = NULL;
  list = NULL;
  hint_need_init = TRUE;
}
